#	include	"Simulator.h"

int
DisembodiedSnoopyCache	::	IsContainedIn(
	SimAddress	sa)
	{
	CacheLine	*cl = theCache->IsInCache(sa);
	if(!cl)			return(NotInCache);
	if(cl->Dirty)	return(InCacheExclusive);
	return(InCacheReadOnly);
	}

bool
DisembodiedSnoopyCache	::	SwitchToReadOnlyAccess(
	SimAddress	sa)
	{
	int		Temp1, Temp2, Temp3;
	// Schedule an SRAM access
	SRAM.RequestNewSchedule(0, AccessLatency, 0, &Temp1, &Temp3);

	CacheLine	*cl = theCache->IsInCache(sa);

	if(!cl)			return(false);
	if(!cl->Dirty)	return(false);

	// Write it back out, but do not flush from cache, just make it
	// no longer dirty
	SnoopyControlUnit	*scu = (SnoopyControlUnit*)AttachedMemorySystem;
	scu->ScheduleSnoopyWriteOperation(
		this, cl->RealAddress<<theCache->ShiftAmount,
		theCache->LineSize, Temp3, &Temp1, &Temp2);
	++theCache->WriteBacks;
	cl->Dirty = 0;	// switch entry to no longer be exclusive
	return(true);
	}

bool
DisembodiedSnoopyCache	::	FlushFromCache(
	SimAddress	sa)
	{
	int		Temp1, Temp2;
	// Schedule an SRAM access
	SRAM.RequestNewSchedule(0, AccessLatency, 0, &Temp1, &Temp2);

	CacheLine	*cl = theCache->IsInCache(sa);

	if(!cl)	return(false);
	if(!cl->Dirty)
		{
		cl->Valid = 0;	// invalidate entry
		return(false);
		}
	SwitchToReadOnlyAccess(sa);
	cl->Valid = 0;	// invalidate entry
	return(true);
	}

bool
DisembodiedSnoopyCache	::	UpdateCache(
	bool		IsWrite,
	SimAddress	sa,
	int			Bytes)
	{
	CacheLine	*cl = theCache->IsInCache(sa);

	if(cl)
		{
		if(IsWrite)
			{
			if(cl->Dirty)
				{
				return(DisembodiedCache::UpdateCache(IsWrite, sa, Bytes));
				}
			}
		else
			{
			return(DisembodiedCache::UpdateCache(IsWrite, sa, Bytes));
			}
		}
	SnoopyControlUnit	*scu = (SnoopyControlUnit*)
			AttachedMemorySystem;

	if(IsWrite)
		scu->GainExclusiveAccess(this, sa, Bytes);
	else
		scu->GainReadOnlyAccess(this, sa, Bytes);

	DisembodiedCache::UpdateCache(IsWrite, sa, Bytes);
	return(true);
	}

