#	include	"Simulator.h"


DisembodiedSnoopyControlUnit	::	DisembodiedSnoopyControlUnit ()
	{
	AttachedMemorySystem = NULL;
	ConnectedCaches = new Array<SnoopyCache *>;
	}

DisembodiedSnoopyControlUnit	::	~DisembodiedSnoopyControlUnit ()
	{
	// nothing for now.  Should delete stuff
	}

bool
DisembodiedSnoopyControlUnit	::	Initialize (
	int		aLineSize,
	int		aFrontSideBusStartupCost,
	int		aFrontSideBusTransferTime)
	{
	BufferedCacheLine = 0;
	BufferedCacheLineValid = 0;
	FrontSideBusBusy = 0;
	FrontSideBusStartupCost = aFrontSideBusStartupCost;
	FrontSideBusTransferTime = aFrontSideBusTransferTime;
	LineSize = aLineSize;
	int		Index1, Index2;
	Index1 = 1;
	Index2 = 0;
	while(Index1<LineSize)
		{
		Index1 = Index1 << 1;
		++Index2;
		}
	// shift2 now contains the "shift amount" which is how much
	// to shift things down to mask off the linesize.  So say
	// linesize is 1, then we don't shift at all.
	ShiftAmount = Index2;
	FrontSideBusBusyCycles = FrontSideBusIdleCycles = ExclusiveAccessRequests =
		ReadOnlyAccessRequests = 0;
	return ( true );
	}

bool
DisembodiedSnoopyControlUnit	::	PrintConfiguration(
	int     Format,
	FILE    *fp)
	{
	PrintValue(Format, fp, "LineSize", LineSize);
	PrintValue(Format, fp, "FrontSideBusStartupCost", FrontSideBusStartupCost);
	PrintValue(Format, fp, "FrontSideBusTransferTime",FrontSideBusTransferTime);
	return ( true );
	}

bool
DisembodiedSnoopyControlUnit    ::  PrintStatistics(
	int     Format,
	FILE    *fp)
	{
	PrintValue(Format, fp, "NumberOfConnectedCPUs", ConnectedCaches->GetSize());
	PrintValue(Format, fp, "FrontSideBusBusyCycles", FrontSideBusBusyCycles);
	PrintValue(Format, fp, "FrontSideBusIdleCycles", FrontSideBusIdleCycles);
	PrintPercent(Format, fp, "FrontSideBusUtilization",
		FrontSideBusBusyCycles, FrontSideBusIdleCycles);
	PrintValue(Format, fp, "ExclusiveAccessRequests", ExclusiveAccessRequests);
	PrintValue(Format, fp, "ReadOnlyAccessRequests", ReadOnlyAccessRequests);
	return( true );
	}
 
bool
DisembodiedSnoopyControlUnit	::	ConnectMemorySystem(
	MemorySystem	*ms)
	{
	AttachedMemorySystem = ms;
	return(true);
	}

bool
DisembodiedSnoopyControlUnit	::	AddConnectedCache(
	SnoopyCache	*c)
	{
	ConnectedCaches->Add((SnoopyCache*)c);
	return(true);
	}

bool
DisembodiedSnoopyControlUnit	::	GainExclusiveAccess(
	SnoopyCache	*From,
	SimAddress	sa,
	int			Bytes)
	{
	// the cache From wants exclusive access.  Note that
	// it does not ask us to get this unless it really
	// needs it (i.e. doesn't already have it).
	MustWriteToMemory = false;
	SimAddress	sa1 = (sa>>ShiftAmount)<<ShiftAmount;


	++ExclusiveAccessRequests;	
	if(sa != sa1)
		Bytes += (sa - sa1);
	
	int	Index = 0;
	int		DoneBytes = 0;
	while(DoneBytes < Bytes)
		{
		while(Index<ConnectedCaches->GetSize())
			{
			SnoopyCache	*c = ConnectedCaches->ElementAt(Index);

			if(c==From)
				{
				++Index;
				continue;
				}
			switch(c->IsContainedIn(sa1))
				{
				case	NotInCache:	break;
				case	InCacheReadOnly:
					c->FlushFromCache(sa1);
					break;
				case	InCacheExclusive:
					c->FlushFromCache(sa1);
					break;
				}
			++Index;
			}
		sa1 += LineSize;
		DoneBytes += LineSize;
		}
	return(true);
	}

bool
DisembodiedSnoopyControlUnit	::	GainReadOnlyAccess(
	SnoopyCache	*From,
	SimAddress	sa,
	int			Bytes)
	{
	SimAddress	sa1 = (sa>>ShiftAmount)<<ShiftAmount;

	MustWriteToMemory = true;	
	++ReadOnlyAccessRequests;	
	if(sa != sa1)
		Bytes += (sa - sa1);
	
	int	Index = 0;
	int		DoneBytes = 0;
	while(DoneBytes < Bytes)
		{
		while(Index<ConnectedCaches->GetSize())
			{
			SnoopyCache	*c = ConnectedCaches->ElementAt(Index);

			if(c==From)
				{
				++Index;
				continue;
				}
			switch(c->IsContainedIn(sa1))
				{
				case	NotInCache:	break;
				case	InCacheReadOnly: break;
				case	InCacheExclusive:
					c->SwitchToReadOnlyAccess(sa1);
					break;
				}
			++Index;
			}
		sa1 += LineSize;
		DoneBytes += LineSize;
		}
		
	return(true);
	}


bool
DisembodiedSnoopyControlUnit	::	NotifyAboutStall(
	Component	*From,
	int			Cycles)
	{
	// ignore stall messages, since caches will use the ScheduleOperation()
	// interface.
	return(true);
	}

int
DisembodiedSnoopyControlUnit	::	SimulateOneClockCycle()
	{
	if(FrontSideBus.Busy())
		{
		--FrontSideBusBusy;
		++FrontSideBusBusyCycles;
		}
	else
		++FrontSideBusIdleCycles;
			
	return(1);	// call back in 1 cycle
	}

bool
DisembodiedSnoopyControlUnit	::	ScheduleSnoopyWriteOperation(
	SnoopyCache	*From,
	SimAddress	sa,
	int			Bytes,
	int			StartAfter,
	int			*UntilStart,
	int			*UntilFinish)
	{
	// SNOOPY writes come in when the protocol mechanisms (get access).
	// Due to the protocol mechanisms, we must also post these out to
	// memory (if the write comes from a flush due to a read-access
	FrontSideBusBusy = StartAfter;
	BackSideBusBusy = 0;
	(*UntilStart) = FrontSideBusBusy;
	
	int	Temp1, Temp2;

	FrontSideBus.RequestNewSchedule(FrontSideBusBusy, FrontSideBusStartupCost +
		FrontSideBusTransferTime*LineSize/8, 0, &Temp1, &Temp2);
	FrontSideBusBusy = Temp2;

	if(MustWriteToMemory)
		{
		AttachedMemorySystem->ScheduleOperation(
			true, ((sa>>ShiftAmount)<<ShiftAmount),
			LineSize, FrontSideBusBusy, &Temp1, &Temp2);
		FrontSideBusBusy = Temp2;
		}

	(*UntilFinish) = FrontSideBusBusy;

	BufferedCacheLine=((sa>>ShiftAmount)<<ShiftAmount);
	BufferedCacheLineValid = 1;
	return(true);	
	}

bool
DisembodiedSnoopyControlUnit	::	ScheduleOperation(
	bool	IsWrite,
	SimAddress	sa,
	int			Bytes,
	int			StartAfter,
	int			*UntilStart,
	int			*UntilFinish)
	{
	FrontSideBusBusy = StartAfter;
	BackSideBusBusy = 0;

	(*UntilStart) = FrontSideBusBusy;

	if(BufferedCacheLineValid)
	if(((SimAddress)(BufferedCacheLine>>ShiftAmount))==((SimAddress)(sa>>ShiftAmount)))
		{
		// "no time at all".  Seems strange, but this is a "snoopy"
		// protocol, so this cache-line buffer that we've been talking
		// about the whole time is actually located in the caches,
		// not with us, so the cache actually already has this
		// cache-line (after fsb)
		(*UntilFinish) = FrontSideBusBusy;
		if(IsWrite)
			BufferedCacheLineValid = 0;
		return(true);	
		}
	// otherwise, now must fetch from outside world
	int	Temp1, Temp2, Temp3;
	if(IsWrite)
		{
		// must get the data from FSB to us
		FrontSideBus.RequestNewSchedule(FrontSideBusBusy,
			FrontSideBusStartupCost + FrontSideBusTransferTime*Bytes/8,
			0, &Temp1, &Temp2);
		FrontSideBusBusy = Temp2;
		}
	else
		{
		// must get read request to us
		FrontSideBus.RequestNewSchedule(FrontSideBusBusy,
			FrontSideBusStartupCost,
			0, &Temp1, &Temp2);
		FrontSideBusBusy = Temp2;
		}

	AttachedMemorySystem->ScheduleOperation(
		IsWrite, ((sa>>ShiftAmount)<<ShiftAmount),
		LineSize, FrontSideBusBusy, &Temp1, &Temp3);

	if(!IsWrite)
		{
		// a read, so must send back the data
		FrontSideBus.RequestNewSchedule(Temp3,
			FrontSideBusStartupCost + FrontSideBusTransferTime*Bytes/8,
			0, &Temp1, &Temp2);
		FrontSideBusBusy = Temp2;
		}

	(*UntilFinish) = FrontSideBusBusy;
	BufferedCacheLine=((sa>>ShiftAmount)<<ShiftAmount);
	BufferedCacheLineValid = 1;
	return(true);
	}

	
