#	include	"Simulator.h"

Cache	*
Cache	::	CreateCache(
	uint	LineSize,
	uint	BankSize,
	uint	Assoc,
	uint	Policy)
	{
	Cache	*theCache;
	uint		Index1, Index2;

	theCache = new Cache;
	theCache->NumberOfCacheBanks = Assoc;
	theCache->Hits = 0;
	theCache->Misses = 0;
	theCache->WriteBacks = 0;
	theCache->ReplacementPolicy = Policy;
	theCache->CacheBanks = (CacheBank*)new CacheBank[Assoc];
	theCache->LineSize = LineSize;
	theCache->LastBankAccess = 0;
	Index1 = 1;
	Index2 = 0;
	while(Index1 < LineSize)
		{
		++Index2;
		Index1 = Index1 << 1;
		}
	theCache->ShiftAmount = Index2;
	Index1 = 1;
	Index2 = 0;
	while(Index1 < BankSize)
		{
		Index2 = (Index2 << 1)|1;
		Index1 = Index1 << 1;
		}
	theCache->EntryMask = Index2;	
	Index1 = 0;
	while(Index1 < Assoc)
		{
		theCache->CacheBanks[Index1].NumberOfCacheLines = BankSize;
		theCache->CacheBanks[Index1].CacheLines =
			(CacheLine*) new CacheLine[BankSize];
		Index2 = 0;
		while(Index2 < BankSize)
			{
			theCache->CacheBanks[Index1].CacheLines[Index2].RealAddress = 0;
			theCache->CacheBanks[Index1].CacheLines[Index2].Dirty = 0;
			theCache->CacheBanks[Index1].CacheLines[Index2].Valid = 0;
			theCache->CacheBanks[Index1].CacheLines[Index2].PolicyField =Index1;
			theCache->CacheBanks[Index1].CacheLines[Index2].SharedField = 0;
			++Index2;
			}
		++Index1;
		}
	return(theCache);
	}

void
Cache	::	UpdatePolicy(
	int				Bank,
	uint			Entry)
	{
	int		OldEntry;
	int		Index1;

	OldEntry = CacheBanks[Bank].CacheLines[Entry].PolicyField;

	Index1 = 0;
	while ( Index1 < NumberOfCacheBanks)
		{
		if(Index1 == Bank)
			CacheBanks[Index1].CacheLines[Entry].PolicyField =
				NumberOfCacheBanks-1;
		else
		if(CacheBanks[Index1].CacheLines[Entry].PolicyField > OldEntry)
			--CacheBanks[Index1].CacheLines[Entry].PolicyField;
		++Index1;
		}
	}


CacheLine	*
Cache	::	LocateInCache(
	uint			aAddress,
	uint			*HitOrMiss,
	uint			*ForcedWriteBack,
	CacheLine		*WrittenCacheLine)
	{
	unsigned	int	Entry;
	unsigned	int	theLine;
	int			Index1, Invalid = -1, Lowest;

	aAddress = aAddress & 0x7ffffffc;
	Entry = (aAddress >> ShiftAmount)&EntryMask;
	theLine = (aAddress >> ShiftAmount);
	(*ForcedWriteBack) = 0;

	// LastBankAccess is an optimization.  Caches exploit
	// locality to improve performance.  Turns out we can
	// do the same thing with cache simulators.
	if(CacheBanks[LastBankAccess].
		CacheLines[Entry].RealAddress == theLine)
	if(CacheBanks[LastBankAccess].CacheLines[Entry].Valid)
		{
		Index1 = LastBankAccess;
		if(CacheBanks[Index1].CacheLines[Entry].PolicyField !=
			(NumberOfCacheBanks-1))
			UpdatePolicy(Index1, Entry);
		++Hits;
		(*HitOrMiss) = 1;
		return(&CacheBanks[Index1].CacheLines[Entry]);
		}
	
	Index1 = 0;
	while(Index1 < NumberOfCacheBanks)
		{
		if(CacheBanks[Index1].CacheLines[Entry].RealAddress == theLine)
		if(CacheBanks[Index1].CacheLines[Entry].Valid)
			{
			if(CacheBanks[Index1].CacheLines[Entry].PolicyField !=
				(NumberOfCacheBanks-1))
				UpdatePolicy(Index1, Entry);
			++Hits;
			(*HitOrMiss) = 1;
			LastBankAccess = Index1;
			return(&CacheBanks[Index1].CacheLines[Entry]);
			}
		if(!CacheBanks[Index1].CacheLines[Entry].Valid)
			Invalid = Index1;
		if(!CacheBanks[Index1].CacheLines[Entry].PolicyField)
			Lowest = Index1;
		++Index1;
		}
	// Not in the cache!	
	++Misses;
	(*HitOrMiss) = 0;

	if(Invalid>=0)
		Index1 = Invalid;
	else
	if(ReplacementPolicy == 0)	// LRU
		Index1 = Lowest;
	else
	if(ReplacementPolicy == 1)	// Random
		Index1 = rand()%NumberOfCacheBanks;

	if(CacheBanks[Index1].CacheLines[Entry].Dirty)
		{
		(*ForcedWriteBack) = 1;
		memcpy(WrittenCacheLine, &CacheBanks[Index1].CacheLines[Entry],
			sizeof(CacheLine));
		++WriteBacks;
		}

	CacheBanks[Index1].CacheLines[Entry].Dirty = 0;
	CacheBanks[Index1].CacheLines[Entry].Valid = 1;
	CacheBanks[Index1].CacheLines[Entry].RealAddress = theLine;
	if(CacheBanks[Index1].CacheLines[Entry].PolicyField !=
		(NumberOfCacheBanks-1))
		UpdatePolicy(Index1, Entry);
	LastBankAccess = Index1;
	return(&CacheBanks[Index1].CacheLines[Entry]);
	}

CacheLine	*
Cache	::	IsInCache(
	uint	aAddress)
	{
	unsigned	int	Entry;
	unsigned	int	theLine;
	int			Index1, Invalid = -1, Lowest;

	aAddress = aAddress & 0x7ffffffc;
	Entry = (aAddress >> ShiftAmount)&EntryMask;
	theLine = (aAddress >> ShiftAmount);

	if(CacheBanks[LastBankAccess].
		CacheLines[Entry].RealAddress == theLine)
	if(CacheBanks[LastBankAccess].CacheLines[Entry].Valid)
		{
		Index1 = LastBankAccess;
		return(&CacheBanks[Index1].CacheLines[Entry]);
		}
	
	Index1 = 0;
	while(Index1 < NumberOfCacheBanks)
		{
		if(CacheBanks[Index1].CacheLines[Entry].RealAddress == theLine)
		if(CacheBanks[Index1].CacheLines[Entry].Valid)
			{
			LastBankAccess = Index1;
			return(&CacheBanks[Index1].CacheLines[Entry]);
			}
		++Index1;
		}
	return(NULL);
	}

